import sys
from PyQt5.QtWidgets import QMainWindow, QAction, qApp, QTextEdit, QWidget
from PyQt5.QtWidgets import QPushButton, QAbstractItemView, QMessageBox, QTableView
from PyQt5.QtGui import QIcon, QStandardItemModel, QStandardItem, QColor, QBrush
from PyQt5.QtCore import pyqtSlot, QEvent, Qt
from add_contact import AddContactDialog

sys.path.append('../')


class MainApp(QMainWindow):
    """
    Описание класса основного окна клиентской программы
    """
    def __init__(self, client_engine, client_db, parent=None):
        super().__init__()
        self.client_engine = client_engine
        self.client_db = client_db

        # Связка между сигналом и методом экземпляра движка?
        self.client_engine.new_user.connect(self.new_user)
        self.client_engine.user_not_exist.connect(self.user_not_exist)
        self.client_engine.new_message.connect(self.new_message)
        self.client_engine.update_users_status.connect(self.update_users_status)

        # Текущий пользователь в списке известных пользователей (текущий чат)
        self.current_user = None
        self.current_user_id = 0

        self.initUI()

    def initUI(self):
        self.setFixedSize(800, 630)
        self.setWindowTitle('Client chat')

        self.known_users_model = None
        self.tv_known_users = QTableView(self)
        self.tv_known_users.move(10, 50)
        self.tv_known_users.setFixedSize(160, 540)
        self.tv_known_users.clicked.connect(self.select_active_user)

        # Устанавливаем способ выделения внутри таблицы: вся строка
        self.tv_known_users.setSelectionBehavior(QAbstractItemView.SelectRows)

        self.messages_model = None
        self.tv_messages = QTableView(self)
        self.tv_messages.move(180, 50)
        self.tv_messages.setFixedSize(610, 400)

        self.te_message = QTextEdit(self)
        self.te_message.move(180, 460)
        self.te_message.setFixedSize(610, 80)

        self.btn_send = QPushButton(self)
        self.btn_send.move(690, 550)
        self.btn_send.setFixedSize(100, 40)
        self.btn_send.setText('Отправить')
        self.btn_send.clicked.connect(self.send_message)

        exitAction = QAction(QIcon(''), 'Завершить работу', self)
        exitAction.setShortcut('Ctrl+Q')
        # exitAction.triggered.connect(qApp.quit)
        # TODO: есть предопределенное собитие в def event() и триггер. Что-то здесь лишнее
        exitAction.triggered.connect(self.close_app)

        self.btnRefreshKnownUsers = QAction(QIcon(''), 'Refresh', self)
        self.btnRefreshKnownUsers.triggered.connect(self.refresh_users_status)

        self.btnAppendUser = QAction(QIcon(''), 'Append user', self)
        self.btnAppendUser.triggered.connect(self.append_user)

        self.btnDeleteUser = QAction(QIcon(''), 'Delete user', self)
        self.btnDeleteUser.triggered.connect(self.delete_user)

        # toolbar
        self.toolbar = self.addToolBar('Toolbar')
        self.toolbar.setFixedHeight(40)
        self.toolbar.addAction(exitAction)
        self.toolbar.addSeparator()
        self.toolbar.addAction(self.btnRefreshKnownUsers)
        self.toolbar.addAction(self.btnAppendUser)
        self.toolbar.addAction(self.btnDeleteUser)

        # statusbar
        self.statusBar()

        # Подгружаем из БД список известных пользователей
        self.show_known_users()

        self.show()

    def refresh_window(self):
        """
        Метод, обновляющий основное окно программы
        Здесь включаются/выключаются поля в зависимости от того, выбран ли текущий пользователь
        :return:
        """
        if self.current_user is None:
            self.btnDeleteUser.setDisabled(True)
            self.tv_messages.setDisabled(True)
            self.te_message.setDisabled(True)
            self.btn_send.setDisabled(True)
        else:
            self.btnDeleteUser.setDisabled(False)
            self.tv_messages.setDisabled(False)
            self.te_message.setDisabled(False)
            self.btn_send.setDisabled(False)

        self.statusBar().showMessage(f'Текущий пользователь: {self.client_engine.username}, '
                                     f'текущий чат: {self.current_user}')

    def refresh_users_status(self):
        """
        Метод, обновляющий статус известных клиентов (online/offline)
        Формирует запрос со списком клиентов и отправляет его на сервер
        :return:
        """
        loc_userslist = []
        for item in self.client_db.session.query(self.client_db.KnownUsers).all():
            loc_userslist.append(item.name)

        # Отправляем запрос на сервер о статусе клиентов
        # Затем из полученных сообщений расшифруем статус и обновим таблицу
        self.client_engine.get_users_status(','.join(loc_userslist))

    def show_known_users(self):
        """
        Метод, описывающий список известных клиентов
        :return:
        """
        self.known_users_model = QStandardItemModel(self)

        for item in self.client_db.session.query(self.client_db.KnownUsers).all():
            table_item = []
            table_item.append(QStandardItem(str(item.id)))
            table_item.append(QStandardItem(item.name))
            if item.name in self.client_engine.online_userslist:
                t_item = QStandardItem('online')
                t_item.setForeground(QBrush(QColor('darkgreen')))
                table_item.append(t_item)
            else:
                table_item.append(QStandardItem('offline'))
                # Потом здесь будет семафорчик, зеленый - online, красный - offline

            # Вот тут бред, setEditable устанавливать на каждую ячейку. Почему нет setEditable для всего data_model?
            # Или на худой конец - для отдельной строки или колонки?
            for item in table_item:
                item.setEditable(False)
            self.known_users_model.appendRow(table_item)

        # Связываем модель с таблицей
        self.tv_known_users.setModel(self.known_users_model)

        # Прячем вертикальные и горизонтальные заголовки
        self.tv_known_users.horizontalHeader().hide()
        self.tv_known_users.verticalHeader().hide()

        # Прячем колонку с id
        self.tv_known_users.setColumnWidth(0, 0)

        # Колонку с username растягиваем на всю ширину
        self.tv_known_users.setColumnWidth(1, 100)
        self.tv_known_users.setColumnWidth(2, 60)

        # Прячем сетку
        self.tv_known_users.setShowGrid(False)

        # Обновляем окно
        self.refresh_window()

    def select_active_user(self):
        """
        Метод, срабатывающий при щелчке на известном пользователе
        Обновляет список сообщений и окно программы
        :return:
        """
        self.current_user = self.tv_known_users.currentIndex().data()
        self.current_user_id = self.client_db.get_user_id(self.current_user)
        self.show_messages()
        self.refresh_window()

    def append_user(self):
        """
        Метод, вызывающий окно для воода имени пользователя для добавления в список известных пользователей
        :return:
        """
        global select_dialog
        select_dialog = AddContactDialog()
        select_dialog.btn_find.clicked.connect(lambda: self.add_contact_action(select_dialog))
        select_dialog.show()

    def add_contact_action(self, item):
        """
        Метод - обработчик добавления, сообщает серверу имя пользователя, обновляет таблицу и список контактов
        :param item - объект диалогового окна:
        :return:
        """
        # Берем из диалога имя пользователя
        new_contact = item.username_edit.text()

        # Вызываем метод, который отправит сообщение на сервер - проверить существование такого пользователя
        self.client_engine.check_username(new_contact)

        # Закрываем диалоговое окно
        item.close()

    def delete_user(self):
        """
        Метод удаляет пользователя из списка известных пользователей
        :return:
        """
        if self.current_user and \
                QMessageBox.question(self, 'Подтвердите удаление записи',
                                     f'Действительно удалить из списка известных пользователей {self.current_user}?',
                                     QMessageBox.Yes|QMessageBox.No,
                                     QMessageBox.No) == QMessageBox.Yes:
            self.client_db.delete_known_user(self.current_user)
            self.show_known_users()
            self.current_user = None
            self.current_user_id = 0
            self.refresh_window()

    def show_messages(self):
        """
        Метод обновляет список сообщений при выборе контакта в списке известных пользователей
        :return:
        """
        self.messages_model = QStandardItemModel(self)

        for item in self.client_db.session.query(self.client_db.Messages).filter(self.client_db.Messages.user_id == self.current_user_id).all():
            # TODO: предыдущая строка валится на предупреждение. Нужно исправлять
            # SAWarning: implicitly coercing SELECT object to scalar subquery; please use the
            # .scalar_subquery() method to produce a scalar subquery
            table_item = []
            if item.direction == 1:
                row_item = QStandardItem(f'Входящее, {item.mdatetime}: \n {item.message}')
                row_item.setBackground(QBrush(QColor(255, 213, 213)))
                row_item.setTextAlignment(Qt.AlignLeft)
            else:
                row_item = QStandardItem(f'Исходящее,  {item.mdatetime}: \n {item.message}')
                row_item.setTextAlignment(Qt.AlignRight)
                row_item.setBackground(QBrush(QColor(204, 255, 204)))
            row_item.setEditable(False)
            table_item.append(row_item)

            self.messages_model.appendRow(table_item)

        self.tv_messages.setModel(self.messages_model)

        # Прячем вертикальные и горизонтальные заголовки
        self.tv_messages.horizontalHeader().hide()
        self.tv_messages.verticalHeader().hide()

        # Колонку с username растягиваем на всю ширину
        self.tv_messages.setColumnWidth(0, 610)

        # Растягиваем высоту строк по содержимому
        self.tv_messages.resizeRowsToContents()

        # Обновляем окно
        self.refresh_window()

    def send_message(self):
        """
        Метод, срабатывающий при нажатии на кнопку отправки введенного сообщения
        :return:
        """
        # Сохраняем сообщение в локальной базе данных
        self.client_db.append_message(self.current_user, 2, self.te_message.toPlainText())

        # `Отдаем сообщение движку для отправки в сокет
        self.client_engine.prepare_message_to_send(self.client_engine.username,
                                                   self.current_user, self.te_message.toPlainText())

        # Очищаем поле для ввода сообщения
        self.te_message.clear()

        # Обновляем историю сообщений
        self.show_messages()

        # Обновляем окно
        self.refresh_window()

    def close_app(self):
        """
        Метод закрытия окноа программы
        :return:
        """
        self.client_engine.engine_shutdown()
        qApp.quit()

    def event(self, e):
        if e.type() == QEvent.Close:
            self.close_app()

        # Событие отправляется дальше
        return QWidget.event(self, e)

    @pyqtSlot()
    def new_user(self):
        self.show_known_users()

    @pyqtSlot()
    def user_not_exist(self):
        """
        Слот выводит на экран сообщение об отсутствии пользователя на сервере
        :return:
        """
        QMessageBox.information(self, 'Ошибка!', 'Пользователь с указанным именем на сервере отсутствует',
                                QMessageBox.Ok,
                                QMessageBox.Ok)

    @pyqtSlot()
    def new_message(self):
        """
        Слот срабатывает при поступлении новго сообщения
        :return:
        """
        self.show_known_users()
        self.show_messages()

    @pyqtSlot()
    def update_users_status(self):
        """
        Слот, срабатывающий при обновлении статусо пользователей
        :return:
        """
        self.show_known_users()



import os
import logging
from logging.handlers import TimedRotatingFileHandler

PATH = os.path.dirname(os.path.abspath(__file__))
PATH = os.path.join(PATH, 'client.log')

logger = logging.getLogger('client')

# Формат строки для вывода сообщений
strfmt = '%(asctime)s.%(msecs)03d %(levelname)-10s %(name)-10s %(message)s'

# Устанавливаем формат даты
datefmt = '%d-%m-%Y %H:%M:%S'
formatter = logging.Formatter(fmt=strfmt, datefmt=datefmt)
fh = TimedRotatingFileHandler(PATH, when='M', interval=10, encoding='utf-8')
fh.setLevel(logging.INFO)
fh.setFormatter(formatter)

logger.addHandler(fh)
logger.setLevel(logging.INFO)
logger.propagate = False



import sys
import logging
import traceback as tb


def log(func):
    def wrap(*args, **kwargs):
        if str(sys.argv).find('client') == -1:
            l = logging.getLogger('server')
        else:
            l = logging.getLogger('client')

        call_str = tb.format_stack()[0]
        n_pos = call_str.find('\n')

        l.info(f'Декоратор: обращение к функции {func.__name__} с параметрами {args}, {kwargs}')
        l.info(f'вызвано из функции {call_str[n_pos + 1 ::]}')
        # Здесь наверное можно было еще поиграться с форматированием, но, как минимум, вызвавшую функцию я смог определить

        f = func(*args, **kwargs)
        return f
    return wrap



from PyQt5.QtWidgets import QDialog, QLabel, QPushButton, QLineEdit
from PyQt5.QtCore import Qt


class AddContactDialog(QDialog):
    """
    Диалог ввода имени пользователя для добавления в список известных пользователей
    Максимально облегченный модуль, только поле для ввода имени пользователя.
    Намеренно не делаю выпадающий список, так как, если на сервере количество пользователей > 20,
    выпадающий список будет неудобным для использования
    """
    def __init__(self):
        super().__init__()

        self.setFixedSize(350, 120)
        self.setWindowTitle('Добавление контакта')
        self.setAttribute(Qt.WA_DeleteOnClose)
        self.setModal(True)

        self.username_label = QLabel('Имя пользователя', self)
        self.username_label.setFixedSize(320, 20)
        self.username_label.move(10, 10)

        self.username_edit = QLineEdit(self)
        self.username_edit.setFixedSize(320, 20)
        self.username_edit.move(10, 40)

        self.btn_find = QPushButton('Найти', self)
        self.btn_find.setFixedSize(100, 40)
        self.btn_find.move(130, 70)

        self.btn_cancel = QPushButton('Отмена', self)
        self.btn_cancel.setFixedSize(100, 40)
        self.btn_cancel.move(240, 70)
        self.btn_cancel.clicked.connect(self.close)



from PyQt5.QtWidgets import QDialog, QLabel, QPushButton, QLineEdit, qApp
from PyQt5.QtCore import Qt


# Ввод имени пользователя при старте приложения
class GetUserName(QDialog):
    """
    Класс, описывающий окно для ввода имени пользователя и пароля при старте приложения
    """
    def __init__(self):
        super().__init__()

        self.setFixedSize(350, 190)
        self.setWindowTitle('Авторизация')
        self.setAttribute(Qt.WA_DeleteOnClose)
        self.setModal(True)

        self.username_label = QLabel('Имя пользователя', self)
        self.username_label.move(10, 10)
        self.username_label.setFixedSize(320, 20)

        self.username_edit = QLineEdit(self)
        self.username_edit.move(10, 40)
        self.username_edit.setFixedSize(320, 20)

        self.passwd_label = QLabel('Пароль', self)
        self.passwd_label.move(10, 70)
        self.passwd_label.setFixedSize(320, 20)

        self.passwd_edit = QLineEdit(self)
        self.passwd_edit.setEchoMode(QLineEdit.Password)
        self.passwd_edit.move(10, 100)
        self.passwd_edit.setFixedSize(320, 20)

        self.btn_ok = QPushButton('OK', self)
        self.btn_ok.move(130, 130)
        self.btn_ok.setFixedSize(100, 40)
        self.btn_ok.clicked.connect(self.click)

        self.btn_cancel = QPushButton('Отмена', self)
        self.btn_cancel.move(240, 130)
        self.btn_cancel.setFixedSize(100, 40)
        self.btn_cancel.clicked.connect(qApp.exit)

        # Нужно добавить обработу нажатия клавиши Esc

        self.show()

    # Обработчик кнопки ОК, если поле вводе не пустое, ставим флаг и завершаем приложение.
    def click(self):
        # TODO: по клавише Esc получаем сообщение об ошибке. Нужно искать решение
        # Traceback (most recent call last): File
        # "...client/client.py", line 22, in < module >
        # if username_dlg.ok_pressed: RuntimeError: wrapped C / C + + object of type GetUserName has been deleted
        if self.username_edit.text():
            self.ok_pressed = True
            qApp.exit()



import json
import argparse
import configparser
import sys
import logging
import os
from socket import *


class AppConfig:
    """
    Класс описывает конфигурацию приложения, его клиентской и серверной части
    """
    def __init__(self, ini_filename='settings.ini'):
        path = os.path.dirname(os.path.realpath(__file__))
        settings_filename = os.path.join(path, ini_filename)

        self.chat_config = config = configparser.ConfigParser()
        self.chat_config.read(settings_filename)

        self.client_host = config['CLIENT']['host']
        self.client_port = config['CLIENT']['port']
        self.client_username = config['CLIENT']['username']
        self.server_hosts = config['SERVER']['hosts']
        self.server_port = config['SERVER']['port']
        self.db_filename = config['DATABASE']['db_filename']


def get_server_params(logger):
    """
    Функция считывает из конфигурационного файла параметры серверной части приложения
    :param logger: Логгер для протоколирования работы метода
    :return: hosts, port, db_filename - считанные параметры
    """
    # TODO: сделать методом класса
    config = AppConfig()

    hosts, port, db_filename = None, None, None

    parser = argparse.ArgumentParser(description='server parser')
    parser.add_argument('-a')
    parser.add_argument('-p')
    parser.add_argument('-db_filename')
    args = parser.parse_args()

    hosts = args.a or config.server_hosts
    port = int(args.p) if args.p else int(config.server_port)
    db_filename = args.db_filename or config.db_filename

    if hosts is None or port is None or db_filename is None:
        logging.getLogger(logger).critical(f'Некорректные параметры командной строки: {sys.argv}')
        sys.exit(1)
    else:
        logging.getLogger(logger).info(f'Определены параметры командной строки: host = {hosts}, port = {port}')

    return hosts, port, db_filename


def get_client_params(logger):
    """
    Функция считывает из конфигурационного файла параметры клиентской части приложения
    :param logger: Логгер для протоколирования работы метода
    :return: host, post, username - считанные параметры
    """
    # TODO: сделать методом класса
    config = AppConfig()

    host, port, username = None, None, None

    parser = argparse.ArgumentParser(description='client parser')
    parser.add_argument('--addr')
    parser.add_argument('--port')
    parser.add_argument('--username')
    args = parser.parse_args()

    host = args.addr or config.client_host
    port = args.port or int(config.client_port)
    username = args.username or config.client_username

    if host is None or port is None or username is None:
        logging.getLogger(logger).critical(f'Некорректные параметры командной строки: {sys.argv}')
        sys.exit(1)
    else:
        logging.getLogger(logger).info(f'Определены параметры командной строки: host = {host}, port = {port}, username = {username}')

    return host, port, username


def encode_msg(msg):
    """
    Функция получает на входе сообщение, кодирует его в UTF-8 и сериализует
    :param msg: Сообщение в виде dict (допустимые форматы сообщений необходимо где-то описать)
    :return: Сериализованное кодированное сообщение
    """
    return json.dumps(msg).encode('utf-8')


def decode_msg(msg):
    """
    Функция получает на входе сериализованное кодированное в UTF-8 сообщение,
    десериализует его и декожирует
    :param msg: Сообщение
    :return: Декодированное сообщение в виде dict
    """
    # Столкнулись с проблемой при декодировании пустых сообщений
    try:
        msg = msg.decode('utf-8')
    except Exception as e:
        print(f'Decode message error: {e}')

    res = ''
    if msg:
        try:
            res = json.loads(msg)
        except Exception as e:
            print(f'JSON loads error: {e}. Message: {msg}')

    return res


def create_socket(logger):
    """
    Функция создает сокет для обмена сообщениями
    :param logger: Логгер для протоколирования работы функции
    :return: Созданный сокет
    """
    # TODO: проверить, есть подозрение, что вызов отсюда уже не используется
    s = None
    try:
        s = socket(AF_INET, SOCK_STREAM)  # Создать сокет TCP
    except Exception as e:
        logging.getLogger(logger).error(f'Ошибка при создании сокета {str(e)}')
        sys.exit(1)
    else:
        logging.getLogger(logger).info(f'Сокет успешно создан: {s}')

    return s



from sqlalchemy import create_engine
from sqlalchemy import Table, Column, Integer, DateTime, String, MetaData, ForeignKey
from datetime import datetime
from sqlalchemy.orm import mapper, sessionmaker
import datetime


class ClientDB:
    """
    Класс для описания локальной БД клиента. Состоит из таблиц:
    KnownUsers - список известных пользователей
    Messages - локальное хранилище сообщений
    """
    class KnownUsers:
        def __init__(self, name):
            self.name = name

        def __repr__(self):
            return f'User: {self.name}'

    # Таблица с сообщениями
    class Messages:
        def __init__(self, mdatetime, user_id, direction, message):
            self.mdatetime = mdatetime
            self.user_id = user_id
            self.direction = direction
            self.message = message
            # Тут можно подойти творчески, добавить признаки доставки сообщения, прочтения сообщения,
            # отложенного сообщения, доставки сообщения в строго определенное время, ...

    def __init__(self, user_name):
        self.sql_engine = None
        try:
            self.sql_engine = create_engine(f'sqlite:///client_{user_name}.db3',
                                            pool_recycle=3600,
                                            connect_args={'check_same_thread': False})
        except Exception as e:
            print(e)

        self.metadata = MetaData()

        known_users = Table('known_users', self.metadata,
                            Column('id', Integer, primary_key=True),
                            Column('name', String, nullable=False, unique=True)
                            )

        messages = Table('messages', self.metadata,
                         Column('id', Integer, primary_key=True),
                         Column('mdatetime', DateTime, default=datetime.datetime.now()),
                         Column('user_id', ForeignKey('known_users.id')),
                         Column('direction', Integer),  # In = 1, Out = 2
                         Column('message', String)
                         )

        self.metadata.create_all(self.sql_engine)

        mapper(self.KnownUsers, known_users)
        mapper(self.Messages, messages)

        Session = sessionmaker(bind=self.sql_engine)
        self.session = Session()

    # Добавяляем пользователя в список известных пользователей
    def append_user(self, name_user):
        if not self.session.query(self.KnownUsers).filter_by(name=name_user).count():
            known_users_row = self.KnownUsers(name_user)
            self.session.add(known_users_row)
            self.session.commit()

    # Получить id пользователя по его имени из локальной (!) БД
    def get_user_id(self, name_user):
        """
        Возвращает id пользователя по его имени
        :param name_user: Имя пользователя
        :return: Id пользователя
        """
        return self.session.query(self.KnownUsers.id).filter_by(name=name_user)

    def delete_known_user(self, name_user):
        self.session.query(self.KnownUsers).filter_by(name=name_user).delete()
        self.session.commit()
        # Здесь важный нюанс! Если удаляем из known_users пользователя, удалять ли каскадно его сообщения?
        # Получается, нарушена ссылочная целостность
        # С другой стороны, все эти сообщения есть на сервере и их можно (при улучшении интерфейса) с сервера запросить

    def append_message(self, name_user, direction, message):
        user_id = self.get_user_id(name_user)
        messages_row = self.Messages(datetime.datetime.now(), user_id, direction, message)
        self.session.add(messages_row)
        self.session.commit()



from setuptools import setup, find_packages

setup(name="rtrv_my_chat_client",
      version="1.0.4",
      description="my_chat_client",
      author="RTretyakov",
      author_email="rtretyakov@outlook.com",
      packages=find_packages(),
      install_requires=['PyQt5', 'sqlalchemy', 'pycryptodome', 'pycryptodomex'],
      scripts=['client/client_run'],
      package_data={
            # If any package contains *.txt files, include them:
            "": ["*.ini"],
            # And include any *.dat files found in the "data" subdirectory
            # of the "mypkg" package, also:
            # "mypkg": ["data/*.dat"],
      }
      )



from PyQt5.QtWidgets import QDialog, QPushButton, QLineEdit, QApplication, QLabel , qApp


# ??? а где эта форма используется???

# Стартовый диалог с выбором имени пользователя
class UserNameDialog(QDialog):
    def __init__(self):
        super().__init__()

        # self.ok_pressed = False

        self.setWindowTitle('Авторизация')
        self.setFixedSize(400, 180)

        self.username_label = QLabel('Имя пользователя:', self)
        self.username_label.move(20, 30)
        self.username_label.setFixedSize(200, 20)

        self.username_edit = QLineEdit(self)
        self.username_edit.move(20, 60)
        self.username_edit.setFixedSize(200, 20)

        self.passwd_label = QLabel('Пароль:', self)
        self.passwd_label.move(20, 90)
        self.passwd_label.setFixedSize(200, 20)

        self.passwd_edit = QLineEdit(self)
        self.passwd_edit.setEchoMode(QLineEdit.Password)
        self.passwd_edit.move(20, 120)
        self.passwd_edit.setFixedSize(200, 20)

        self.btn_ok = QPushButton('Начать', self)
        self.btn_ok.move(200, 150)
        self.btn_ok.setFixedSize(80, 30)
        self.btn_ok.clicked.connect(self.click)

        self.btn_cancel = QPushButton('Выход', self)
        self.btn_cancel.move(300, 150)
        self.btn_cancel.setFixedSize(80, 30)
        self.btn_cancel.clicked.connect(qApp.exit)

        self.show()

    # Обработчик кнопки ОК, если поле вводе не пустое, ставим флаг и завершаем приложение.
    def click(self):
        if self.client_name.text():
            self.ok_pressed = True
            qApp.exit()


if __name__ == '__main__':
    app = QApplication([])
    dial = UserNameDialog()
    app.exec_()



import sys
from service.service import get_client_params
from PyQt5.QtWidgets import QApplication
from client_db import ClientDB
from client_engine import ClientEngine
from client_gui import MainApp
from get_username import GetUserName


# Определяем параметры командной строки
host, port, username = get_client_params(sys.argv, 'client')

# Стартуем приложение
app = QApplication(sys.argv)

# Запрашиваем имя пользователя, если его не было в параметрах командной строки
if not username:
    username_dlg = GetUserName()
    app.exec_()

    # Здесь неадекватная реакция при нажатии клавиши Esc
    if username_dlg.ok_pressed:
        username = username_dlg.username_edit.text()
        userpasswd = username_dlg.passwd_edit.text()
        del username_dlg
    else:
        exit(0)

# Для пользователя подключаем его БД
client_db = ClientDB(username)

# Клиентский движок
client_engine = ClientEngine(host, port, client_db, username, userpasswd)
if client_engine.user_authenticate:
    client_engine.setDaemon(True)
    client_engine.start()

    main_app = MainApp(client_engine, client_db)
    sys.exit(app.exec_())

    client_engine.engine_shutdown()
    client_engine.join()
else:
    print('Ошибка аутентификации пользователя!')



import sys
import hmac
import threading
import time
import hashlib
import binascii
from socket import *
from PyQt5.QtCore import pyqtSignal, QObject
from service.service import encode_msg, decode_msg
from decos import log
import logging


socket_lock = threading.Lock()
SECRET_KEY = b'abrakadabra'


class ClientEngine(threading.Thread, QObject):
    """
    Класс для описания клиентского движка, отвечающего за работу клиентской части
    """
    # Здесь описания сигналов, по которым будут генериться реакции на события
    new_user = pyqtSignal()
    user_not_exist = pyqtSignal()
    new_message = pyqtSignal()
    update_users_status = pyqtSignal()

    def __init__(self, host, port, client_db, username, userpasswd):
        threading.Thread.__init__(self)
        QObject.__init__(self)

        self.host = host
        self.port = port
        self.client_db = client_db
        self.username = username
        self.userpasswd = userpasswd
        self.user_id = self.client_db.get_user_id(self.username)
        self.online_userslist = []
        self.user_authenticate = False

        self.logger = None
        self.logger_name = 'client'
        self.get_logger()

        # Создаем сокет
        self.socket = None
        self.create_socket()
        # Устанавливаем соединение с свервером и проходим процедуру авторизации и аутентификации
        self.make_connection()
        #  Проходим аутентификацию
        if self.user_authenticate:
            self.register_user_on_server()
            self.running = True

    @log
    def get_logger(self):
        try:
            self.logger = logging.getLogger(self.logger_name)
        except Exception as e:
            print(f'Возникла ошибка при создании логгера {str(e)}')
            sys.exit(1)
        else:
            self.logger.info(f'Логгер успешно создан: {self.logger}')

    @log
    def create_socket(self):
        try:
            self.socket = socket(AF_INET, SOCK_STREAM)  # Создать сокет TCP
        except Exception as e:
            self.logger.error(f'Ошибка при создании сокета {str(e)}')
            sys.exit(1)
        else:
            self.logger.info(f'Сокет успешно создан: {self.socket}')

    @log
    def make_connection(self):
        try:
            self.socket.connect((self.host, self.port))
        except Exception as e:
            self.logger.error(f'Возникла ошибка при установке соединения с сервером {str(e)}')
            sys.exit(1)

        self.logger.info(f'Авторизация пользователя {self.username} на сервере')

        # Извещаем сервер о желании авторизоваться. Здесь общаемся с сервером напрямую, без send_message()
        # Получаем тестовое сообщение от сервера
        message = self.socket.recv(32)
        # Вычисляем хэш
        hash = hmac.new(SECRET_KEY, message, 'MD5')
        digest = hash.digest()
        # Отправляем сгенерированный хэш серверу для сравнения
        self.socket.send(digest)

        # Получаем ответ от сервера
        message = self.socket.recv(32).decode('utf-8')
        if message == 'Auth ok!':
            self.logger.info(f'Авторизация пользователя {self.username} на сервере прошла успешно!')

            # Сразу же проводим аутентификацию
            # Пароль пересылаем в зашифрованном (!) виде
            userpasswd = self.userpasswd.encode('utf-8')
            salt = self.username.lower().encode('utf-8')
            passwd_hash = hashlib.pbkdf2_hmac('sha512', userpasswd, salt, 100000)
            passwd_hash_string = binascii.hexlify(passwd_hash).decode('ascii')

            msg = encode_msg({'username': self.username, 'passwd': passwd_hash_string})
            self.socket.send(msg)
            msg = self.socket.recv(1024).decode('utf-8')
            # Потенциальная уязвимость: сообщение возвращается в незашифрованном виде, возможна подмена сообщения!
            if msg == 'Auth ok!':
                self.logger.info(f'Аутентификация пользователя {self.username}!')
                self.user_authenticate = True
            else:
                self.logger.info(f'Ошибка при аутентификации пользователя {self.username}!')

        else:
            self.logger.info(f'Авторизация пользователя {self.username} на сервере завершилась неудачей!')

    @log
    def send_message(self, msg):
        # Принудительно делаем секундную паузу, так как при старте клиента в сокет летят сразу два сообщения
        # и сервер падает. Вероятно, нужно делать очередь на отправку сообщений по аналогии с responses у сервера
        # time.sleep(1)
        with socket_lock:
            self.socket.settimeout(0.5)
            if not self.socket.send(msg):
                self.logger.error('Ошибка при отправке сообщения на сервер! Работа программы будет зваершена')
                sys.exit(1)
            # self.receive_answer()
            # self.process_server_ans(get_message(self.transport))
            self.logger.info(f'Отправлено сообщение на сервер')

    @log
    def register_user_on_server(self):
        self.logger.info(f'Регистрация пользователя {self.username} на сервере')
        msg = encode_msg({'action': 'append_user', 'from_user': self.username, 'to_user': 'server'})
        self.send_message(msg)

    @log
    def prepare_message_to_send(self, from_user, to_user, message):
        self.logger.info(f'Отправка сообщения пользователю {to_user}')
        msg = encode_msg({'action': 'message', 'from_user': from_user, 'to_user': to_user, 'message': message})
        self.send_message(msg)

    @log
    def read_message(self):
        resp = None
        try:
            resp = self.socket.recv(1024)
        except Exception:
            pass
        else:
            if resp:
                resp = decode_msg(resp)
                self.logger.info(resp)

        return resp

    @log
    def check_username(self, username):
        # Вызывается при добавлении нового пользователя
        self.logger.info(f'Проверка существования пользователя')
        msg = encode_msg({'action': 'check_username', 'from_user': self.username,
                          'to_user': 'server', 'username': username})
        self.send_message(msg)

    @log
    def get_users_status(self, userslist):
        self.logger.info(f'Получение статуса пользователя (онлайн/офлайн)')
        msg = encode_msg({'action': 'get_users_status', 'from_user': self.username,
                          'to_user': 'server', 'userslist': userslist})
        self.send_message(msg)

    @log
    def parse_message(self, msg):
        # Потрошим сообщение
        if msg['action'] == 'check_username':
            # Ответ на запрос проверки существования пользователя
            if msg['message']:
                self.logger.info(f'Успешная проверка пользователя {msg["username"]}')

                # Добавяляем пользователя в локальную БД
                self.client_db.append_user(msg['username'])

                # И генерим событие, по которому обновится список известных пользователей
                self.new_user.emit()
            else:
                # Такого пользователя на сервере не существует
                # Генерим событие, чтобы отобразить messagebox. Отсюда messagebox вызвать не можем
                self.user_not_exist.emit()

        elif msg['action'] == 'get_users_status':
            # Получили ответ на запрос о статусе клиентов
            self.online_userslist = msg['userslist'].split(',')

            # Генерим событие
            self.update_users_status.emit()

        elif msg['action'] == 'auth_user':
            if msg['message'] == 'Auth ok!':
                self.logger.info(f'Успешная аутентификация пользователя {msg["username"]}')
            elif msg['message'] == 'Auth fail!':
                # Генерим событие
                self.update_users_status.emit()

        elif msg['action'] == 'message':
            # Обычное сообщение
            # Проверяем существование и добавляем пользователя в known_users
            self.client_db.append_user(msg['from_user'])

            # Сохраняем сообщение в локальную базу
            self.client_db.append_message(msg['from_user'], 1, msg['message'])

            # Генерим событие, чтобы обновить список сообщений в окне
            self.new_message.emit()

    @log
    def run(self):
        self.logger.info('Запущен процесс - приёмник собщений с сервера.')
        while self.running:
            # Отдыхаем секунду и снова пробуем захватить сокет.
            # если не сделать тут задержку, то отправка может достаточно долго ждать освобождения сокета.
            time.sleep(1)
            with socket_lock:
                self.socket.settimeout(0.5)
                message = self.read_message()
                if message:
                    self.logger.info(f'Принято сообщение с сервера: {message}')
                    self.parse_message(message)

    def engine_shutdown(self):
        self.running = False
        msg = encode_msg({'action': 'kill_user', 'from_user': self.username, 'to_user': 'server'})
        self.send_message(msg)

        self.logger.debug('Транспорт завершает работу.')
        time.sleep(0.5)
